#include <EEPROM.h>
#include "LowPower.h"
#include "datatype.h"

int led = 13;

char print = 0;
char cmd = 0x00;
unsigned char val;

unsigned int status;

int total = 0;
int offset = 0;
unsigned int cont = 0;

_Temp work;
_Temp max;
Temp base;

/*
32
8 8 8 8
8+4 4+6 2+8
12 10 10
*/
void encode(_Temp &in) {
    base.B[0] = in.t >> 4;
    base.B[1] = ((in.t & 0x0F) << 4)| ((in.temp >> 6) & 0x0F);
    base.B[2] = ((in.temp & 0x3F) << 2) | ((in.Vcc >> 8) & 0x03);
    base.B[3] = in.Vcc & 0xFF;
}
void decode(_Temp &out) {
    out.t = 0;
    out.temp = 0;
    out.Vcc = 0;
    out.t = (base.B[0] << 4) | ((base.B[1] & 0xF0) >> 4);
    out.temp = ((base.B[1] & 0x0F) << 6) | ((base.B[2] & 0xFC) >> 2);
    out.Vcc = ((base.B[2] & 0x03) << 8) | base.B[3];
}

void setup()
{
    pinMode(led, OUTPUT);


    EEPROM.get( 0, status );

    total = status;
    offset = sizeof(unsigned int) + (status + 1) * sizeof(Temp);

    EEPROM.get( sizeof(unsigned int), base );
    decode(max);

    if(MCUSR & _BV(EXTRF)){
        Serial.begin(115200);
        print = 1;

        Serial.println( "Read status from EEPROM: " );
        Serial.print( "total\t" );Serial.println( status );
        Serial.print( "max_t\t" );Serial.println( max.t );
        Serial.print( "max_temp\t" );Serial.println( (((long)max.temp*2000 - 648620L) / 144 + 1) >> 1 );
        Serial.print( "max_Vcc\t" );Serial.println( 1126400L / max.Vcc );

        if(status){
            dump();
        }

    }else{
        LowPower.powerDown(SLEEP_2S, ADC_OFF, BOD_OFF); 
        total++;
        rec();
    }
}

// https://code.google.com/p/tinkerit/wiki/SecretVoltmeter
// The voltage is returned in millivolts. So 5000 is 5V, 3300 is 3.3V.
unsigned int vccVoltage() {
    unsigned int result;
    // Read 1.1V reference against AVcc
    ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
    delay(2); // Wait for Vref to settle
    ADCSRA |= _BV(ADSC); // Convert
    while (bit_is_set(ADCSRA,ADSC));
    result = ADCL;
    result |= ADCH << 8;
    // result = 1126400L / result; // Back-calculate AVcc in mV
    return result;
}

// https://code.google.com/p/tinkerit/wiki/SecretThermometer
// Temperature is returned in 10x°C. So 250 is 25°C.
unsigned int readTemp() {
    //long result;
    unsigned int result;
    // Read temperature sensor against 1.1V reference
    ADMUX = _BV(REFS1) | _BV(REFS0) | _BV(MUX3);
    delay(50); // Wait for Vref to settle
    ADCSRA |= _BV(ADSC); // Convert
    while (bit_is_set(ADCSRA,ADSC));
    result = ADCL;
    result |= ADCH << 8;
    //result = (result*1000 - 324310L) / 122;
    //result = ((result*2000 - 648620L) / 144 + 1) >> 1;
    return result;
}

void sleepMinutes()
{
    for (unsigned int i = 0; i < 150; i++) { 
    //for (unsigned int i = 0; i < 1; i++) { 
        LowPower.powerDown(SLEEP_2S, ADC_OFF, BOD_OFF); 
    }
    cont++;
}

void blinkLed(){
    digitalWrite(led, HIGH);
    delay(100);  
    digitalWrite(led, LOW);
}

void rec(){
    work.t = cont;
    work.temp = readTemp();
    work.Vcc = vccVoltage();

    if(work.temp > max.temp){
        max.t = work.t;
        max.temp = work.temp;
        max.Vcc = work.Vcc;
        encode(max);
        EEPROM.put( sizeof(unsigned int), base );
    }
    status = total;

    encode(work);
    EEPROM.put( offset, base );
    EEPROM.put( 0, status );

    total++;
    offset += sizeof(Temp);
    if(offset >= EEPROM.length()){
        total = 0;
        offset = sizeof(unsigned int) + sizeof(Temp);
    }
}

void dump(){
    unsigned int i = 0;

    EEPROM.get( 0, status );

    EEPROM.get( sizeof(unsigned int), base );
    decode(max);
    Serial.print( "max_time\t" );Serial.println( max.t );
    Serial.print( "max_temp\t" );Serial.println( (((long)max.temp*2000 - 648620L) / 144 + 1) >> 1 );
    Serial.print( "max_Vcc\t" );Serial.println( 1126400L / max.Vcc );

    for(offset=sizeof(unsigned int)+sizeof(Temp); offset+sizeof(Temp)<=EEPROM.length(); offset+=sizeof(Temp)){
        EEPROM.get( offset, base);
        decode(work);
        Serial.print( i++ );Serial.print( "\t" );
        //Serial.print( offset );Serial.print( "\t" );
        Serial.print( work.t );Serial.print( "\t" );
        Serial.print( (((long)work.temp*2000 - 648620L) / 144 + 1) >> 1 );Serial.print( "\t" );
        Serial.print( 1126400L / work.Vcc );Serial.println();
    }
}

void loop()
{
    if(print == 0){
        blinkLed();
        sleepMinutes();
        rec();
    }else{
        if(Serial.available()){
            cmd = Serial.read();
            if(cmd == 'C'){
                Serial.print( "clear EEPROM..." );
                status = 0;
                max.t = 0;
                max.temp = 0;
                max.Vcc = 0;
                encode(max);
                EEPROM.put( 0, status );
                EEPROM.put( sizeof(unsigned int), base );
                Serial.println( "done" );
            }
            if(cmd == 'F'){
                Serial.print( "set EEPROM to 0xFF..." );
                for(offset=sizeof(unsigned int)+sizeof(Temp); offset<EEPROM.length(); offset++){
                    EEPROM.write(offset, 0xFF);
                }
                Serial.println( "done" );
            }
            if(cmd == 'D'){
                Serial.println( "Read status from EEPROM: " );
                Serial.print( "total\t" );Serial.println( status );
                Serial.println( "dump EEPROM..." );
                dump();
            }
            if(cmd == 'R'){
                Serial.println( "Raw dump from EEPROM: " );
                for(unsigned int address=0; address<EEPROM.length(); address++){
                    val = EEPROM.read(address);

                    Serial.print(address);
                    Serial.print("\t");
                    Serial.print(val, HEX);
                    Serial.println();
                }
            }
            if(cmd == 'N'){
                Serial.print( "now_temp\t" );Serial.println( (((long)readTemp()*2000 - 648620L) / 144 + 1) >> 1 );
                Serial.print( "now_Vcc\t" );Serial.println( 1126400L / vccVoltage() );
            }
        }
    }
}


